/* FILE: mmdisp.cc                 -*-Mode: c++-*-
 *
 * Mainline for C++/Tcl/Tk 2D vector field display program
 *
 */

#include "machine.h"	/* SYSTEM_TYPE, ... */

// Eventually get build.tcl to find this header file.
#include "oommf.h"	/* Oommf_Init */
#include "misc.h"	/* PlainError */

#include "tclcmds.h"	/* Mmdisp_Init, TclCommandsPreInit, ... */
#include "mmdisp.h"	/* Mmd_Panic, Mmd_WinPanic */

#include <tcl.h>	/* TCL_OK, ... */
#include <tk.h>		/* Tk_Main */
#include <stdio.h>	/* freopen */
#include <stdlib.h>	/* strcmp (?), ... */

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

// There's an annoying bug in the Tk_Init routine of Tk 4.2
// The Tk_Init routine looks in several places for valid Tk initialization 
// scripts.  One place it should look is in ../tk$tk_version relative to the 
// Tcl initialization script library directory.  However, the bug makes it 
// look in ../lib/tk$tk_version instead.  This buggy behavior is masked 
// within wish, since the Tk_Init routine also looks in ../lib/tk$tk_version 
// relative to the wish executable, which happens to be the same directory.  
// The bug surfaces only when another executable links the Tk library, but 
// is not stored in the same directory as the wish executable.
//
// The following brief Tcl script is a workaround of that bug.  It forces the
// proper directory relative to the Tcl script library to be on the list
// of candidate directories searched for Tk initialization scripts.
// This script is sourced within the Tcl_AppInit routine below, between the
// Tcl initialization and the Tk initialization.
//
// See the source code of Tk_Init for more details.
static char tclInitPatch[] =
"if {[string match 7.6 $tcl_version] && ![info exists tk_library]} {\n\
   set tk_library [file join [file dirname $tcl_library] tk4.2]\n\
 }";

int Tcl_AppInit(Tcl_Interp *interp)
{
    if (Tcl_Init(interp) == TCL_ERROR) {
       panic("Tcl initialization error:\n\n%s\n\n\
As a workaround, set the environment variable\n\
TCL_LIBRARY to the name of a directory\n\
containing an error-free init.tcl.\n\n\
For more information see the INSTALL file.", interp->result);
    }
    if (Tcl_Eval(interp, tclInitPatch) != TCL_OK) {
       panic("Tcl/Tk initialization error:\n\n\
Unable to patch installation bug in Tk 4.2\n\n\
Patch script returned error:%s", interp->result);
    }
    if (Tk_Init(interp) == TCL_ERROR) {
       panic("Tk initialization error:\n\n%s\n\n\
As a workaround, set the environment variable\n\
TK_LIBRARY to the name of a directory\n\
containing an error-free tk.tcl.\n\n\
For more information see the INSTALL file.", interp->result);
    }
    Tcl_StaticPackage(interp, "Tk", Tk_Init, (Tcl_PackageInitProc *) NULL);
    if (Oommf_Init(interp) == TCL_ERROR) {
       panic("%s", interp->result);
    }
   if (Mmdisp_Init(interp) == TCL_ERROR) {
       panic("%s", interp->result);
    }
// Initialization successful.
// Now we can use the PlainError routine to handle all fatal error
// conditions, even those generated by routines in the Tcl and Tk
// libraries
    Tcl_SetPanicProc(&Mmd_Panic);
//    panic("Tcl_AppInit Completed!");
    return TCL_OK;
}

#if (SYSTEM_TYPE==WINDOWS)
void
Mmd_WinPanic TCL_VARARGS_DEF(char *,arg1)
{
  va_list argList;
  char buf[4096];
  char *format;

  format = TCL_VARARGS_START(char *,arg1,argList);
  vsprintf(buf, format, argList);

  MessageBeep(MB_ICONEXCLAMATION);
  MessageBox(NULL, buf, "Message from mmDisp",
      MB_ICONSTOP | MB_OK | MB_TASKMODAL | MB_SETFOREGROUND);
  ExitProcess(1);
}
#endif

void
Mmd_Panic TCL_VARARGS_DEF(char *,arg1)
{
  va_list argList;
  char buf[4096];
  char *format;

  format = TCL_VARARGS_START(char *,arg1,argList);
  vsprintf(buf, format, argList);

  PlainError(-1,"%s",buf);

  // Should never get here -- just in case
  exit(1);
}

void Usage()
{
#if (SYSTEM_TYPE==WINDOWS)
  panic("Usage: mmdisp [--server] [--version] [--] [filename]");
  // --nofork flag not ignored under Windows
#else
  panic("Usage: mmdisp [--server] [--nofork] [--version] [--] [filename]");
#endif
  exit(1);
}

void ShowVersion()
{
  panic("mmDisp version: %s\nBuilt against Tcl %s",
	MMDISP_PATCH_LEVEL,TCL_PATCH_LEVEL);
  exit(1);
}

void ParseParams(char *paramstr,int argv_size, int &argc,char** &argv)
{ // The command line does not get parsed (tokenized) under Windows
  // (unlike unix) so this routine provides such to ease portability
  // between Windows and unix.  One NOTE: Under windows, the command
  // name is _NOT_ included in the parameter list!
  // Imports:  argv is a buffer allocated in the client of length
  //   (at least) argv_size.  paramstr is the string holding the
  //   untokenized list.
  // Exports: argc is the number of tokens, which will be <=argv_size,
  //   and argv[] will hold pointers into paramstr to those tokens.
  //   If argc<argv_size, then argv[argc] is set to NULL.
  // Note: This routine modifies paramstr[] (through strtok()).
  const char* whitespace=" \t\n\r\f";
  char *cptr=strtok(paramstr,whitespace);
  argc=0;
  while(argc<argv_size) {
    argv[argc]=cptr;
    if(cptr==NULL) break;
    argc++;
    cptr=strtok(NULL,whitespace);
  }
}

void CheckParams(int argc,char **argv, char* &filename,BOOL &forkflag)
{
  filename=NULL;   // Default is no file
  forkflag=1;      // Default is to fork a child process (if possible)

  int i,flagsdone=0;
  for(i=1;i<argc;i++) {
    if(flagsdone || argv[i][0]!='-') {
      // Filename
      if(filename!=NULL) Usage();   // Only 1 file allowed!
      filename=argv[i];
    }
    else if(strcmp(argv[i],"--nofork")==0) {
      forkflag=0;
    }
    else if(strcmp(argv[i],"--server")==0) {
      // Save value to be given to interpreter later
      Mmd_SaveServerFlag(1);
    }
    else if(strcmp(argv[i],"--version")==0
	    || strcmp(argv[i],"-v")==0) {
      ShowVersion();
    }
    else if(strcmp(argv[i],"--")==0) {
      flagsdone=1;  // No more command line flags.  Anything else on
      /// command line will be interpreted as a filename.
    }
    else {
      Usage();
    }
  }
}

#if (SYSTEM_TYPE==WINDOWS)
int APIENTRY WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,
		     LPSTR lpszCmdLine,int nCmdShow)
{
  // On Windows, use a MessageBox to report initialization errors
  Tcl_SetPanicProc(&Mmd_WinPanic);

  int argc;
  char *argv[32];  // Currently shouldn't need more than 4!
  char **stupid=&(argv[1]); // MS C++ compiler (version 4)
  /// apparently doesn't want to convert from a char *[32]
  /// to a char **& in a parameter list.  But this is okay...
  ParseParams(lpszCmdLine,sizeof(argv)-1,argc,stupid);
  /// NOTE: lpszCmdLine does _not_ include the executable name!
  argc++;
  char exename[128];  // Needs to be long enough to hold "mmdisp".
  strncpy(exename,"mmdisp",sizeof(exename));
  argv[0]=exename;

  char *filename;
  BOOL forkflag;
  CheckParams(argc,argv,filename,forkflag);
  // forkflag ignored under Windows

  TclCommandsPreInit(argv[0],filename);

  argc=1; /* Tk_Main will try to input command line arg's as input files */
  argv[1]="NUL:";
  argc=2;
  Tk_Main(argc,argv,Tcl_AppInit);

panic("Tk_Main returned!");
  // NOTE: I don't know if Windows automatically releases "new"
  //   (heap) memory on program exit.  Since Tk_Main doesn't return,
  //   I can't "delete" any heap memory after this point, so I'll
  //   just hope that argv & exename above get cleaned up by the system.

  return 0;
}
#elif (SYSTEM_TYPE == UNIX)
int main(int argc,char **argv)
{
  char *filename;
  BOOL forkflag;
  CheckParams(argc,argv,filename,forkflag);
  char *fakeargv[32];	// Should only need 2!

#if DEBUGLVL<1
  if(!forkflag || fork()==0) { // Fork only occurs if forkflag!=0
    // Child process
    freopen("/dev/null","r",stdin);
    freopen("/dev/null","r",stdout);
#endif // DEBUGLVL<1
    TclCommandsPreInit(argv[0],filename);
    // Make Tk library think the command line was '<mmdisp> /dev/null'
    fakeargv[0]=argv[0];
    char nullfile[128];  // Needs to be long enough to hold /dev/null
    strncpy(nullfile,"/dev/null",sizeof(nullfile));
    fakeargv[1]=nullfile;
#if DEBUGLVL<1
    int fakeargc=2;
#else
    int fakeargc=1;
#endif // DEBUGLVL<1
    Tk_Main(fakeargc,fakeargv,Tcl_AppInit);
#if DEBUGLVL<1
  }
#endif // DEBUGLVL<1

  return 0;
}
#else
#error "Unknown SYSTEM_TYPE"
#endif // SYSTEM_TYPE == WINDOWS
