/* FILE: mifio.cc           -*-Mode: c++-*-
 *
 * I/O routines involving *.mif (Micromagnetic Input File) files.
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "misc.h"
#include "maginit.h"
#include "mifio.h"
#include "zeeman.h"
#include "maganis.h"

/* End includes */

#define MifSlowPipeName  "SlowPipe"
#define MifFastPipeName  "FastPipe"
#define Mif2dSlabName    "2dSlab"
#define Mif3dSlabName    "3dSlab"
#define Mif3dchargeName  "3dCharge"
#define MifNoDemagName   "NoDemag"

void FieldRangeArray::Free(void)
 {
  if(RangeCount>0) {
    if(StartField!=(double *)NULL) delete[] StartField;
    if(StopField!=(double *)NULL)  delete[] StopField;
    if(StepCount!=(int *)NULL)     delete[] StepCount;
  }
  RangeCount=0;
  StartField=StopField=(double *)NULL;
  StepCount=(int *)NULL;
  RangesFilled=StepTotal=CurIndex=0;
}
int FieldRangeArray::AreSameField(double* field1,double* field2)
{ // Returns 1 if field1==field2, 0 otherwise
  int i;
  for(i=0;i<MIFVECDIM;i++)
    if(field1[i]!=field2[i]) return 0;
  return 1;
}

FieldRangeArray::FieldRangeArray(void)
{
  RangeCount=0;
  StartField=StopField=(double *)NULL;
  StepCount=(int *)NULL;
  RangesFilled=StepTotal=CurIndex=0;
}

void FieldRangeArray::MakeSpace(int newrangecount)
{
  if(newrangecount<1)
    PlainError(1,"Error in FieldRangeArray::MakeSpace: "
	     "Illegal rangecount (%d) request",newrangecount);
  if(newrangecount!=RangeCount) {
    Free(); // Release any previously allocated space
    RangeCount=newrangecount;
    if(!(StartField=new double[RangeCount*MIFVECDIM]) ||
       !(StopField=new double[RangeCount*MIFVECDIM])  ||
       !(StepCount=new int[RangeCount]))
      PlainError(1,"Error in FieldRangeArray::MakeSpace: %s",ErrNoMem);
  }
  RangesFilled=StepTotal=CurIndex=0;
  int i,j;
  for(i=0;i<RangeCount;i++) {
    StepCount[i]=0;
    for(j=0;j<MIFVECDIM;j++)
      StartField[MIFVECDIM*i+j]=StopField[MIFVECDIM*i+j]=0.0;
  }
}

int FieldRangeArray::AddRange(const double* start,const double* stop,
			      int steps)
{
  if(RangesFilled>=RangeCount) {
    PlainWarning("Warning in FieldRangeArray::AddRange: "
		"Attempt to add more ranges than specified; skipping");
    return RangesFilled;
  }
  if(start==(double *)NULL || stop==(double *)NULL || steps<0)
    PlainError(1,"Error in FieldRangeArray: %s",ErrBadParam);

  int i,onefieldflag;
  for(i=0,onefieldflag=1;i<MIFVECDIM;i++) {
    StartField[MIFVECDIM*RangesFilled+i]=start[i];
    StopField[MIFVECDIM*RangesFilled+i] =stop[i];
    if(start[i]!=stop[i]) onefieldflag=0;
  }
  if(onefieldflag) steps=0;  // Degenerate case: single field, no real "range"
  StepTotal+=(StepCount[RangesFilled]=steps);
  if(RangesFilled>0 &&
     !AreSameField(StopField+MIFVECDIM*(RangesFilled-1),
		   StartField+MIFVECDIM*RangesFilled))
    StepTotal+=1; // Step needed from last stop field to this start field

  return ++RangesFilled;
}

int FieldRangeArray::AddRange(const double* onefield)
{
  double start[MIFVECDIM],stop[MIFVECDIM];
  int i,steps;
  for(i=0;i<MIFVECDIM;i++) {
    if(onefield==(double *)NULL) start[i]=stop[i]=0.0;
    else                         start[i]=stop[i]=onefield[i];
  }
  steps=0;
  return AddRange(start,stop,steps);
}

int FieldRangeArray::AddRange(char** param)
{
  double start[MIFVECDIM],stop[MIFVECDIM];
  int i,steps;
  for(i=0;i<MIFVECDIM;i++) {
    start[i]=atof(param[i]);
    stop[i]=atof(param[MIFVECDIM+i]);
  }
  steps=atoi(param[2*MIFVECDIM]);
  return AddRange(start,stop,steps);
}

int FieldRangeArray::ConsistencyCheck(void)
{
  if(RangeCount<1) return 0;  // Nothing to check
  int errcount=0;
  if(RangesFilled<RangeCount) {
    PlainWarning("Warning in FieldRangeArray::ConsistencyCheck: "
		 "Only %d ranges filled out of specified %d",
		 RangesFilled,RangeCount);
    errcount++;
  }
  if(StartField==(double *)NULL || StopField==(double *)NULL
     || StepCount==(int *)NULL) {
    PlainWarning("Warning in FieldRangeArray::ConsistencyCheck: "
		 "Field arrays not initialized");
    return ++errcount;
  }

  int i,tempstep=0;
  for(i=0;i<RangesFilled;i++) {
    tempstep+=StepCount[i];
    if(i>0 && !AreSameField(StopField+MIFVECDIM*(i-1),StartField+MIFVECDIM*i))
      tempstep++;
  }
  if(tempstep!=StepTotal) {
    PlainWarning("Warning in FieldRangeArray::ConsistencyCheck: "
		 "Incorrect value for StepTotal detected (%d instead of %d)",
		 StepTotal,tempstep);
    errcount++;
  }

  if(CurIndex<0 || CurIndex>StepTotal) {
    PlainWarning("Warning in FieldRangeArray::ConsistencyCheck: "
		 "CurIndex (%d) is out-of-bounds [0,%d]",
		 CurIndex,StepTotal);
    errcount++;
  }

  return errcount;
}

int FieldRangeArray::SetFieldIndex(int newindex)
{
  if(newindex<0 || newindex>StepTotal) newindex=StepTotal;
  return CurIndex=newindex;
}

void FieldRangeArray::StepIndex(int& errorcode)
{ // Sets errorcode=0 on success,
  //               -1 if already at last step
  errorcode=0;
  CurIndex++;
  if(CurIndex<0) CurIndex=0;
  else if (CurIndex>StepTotal) {
    CurIndex=StepTotal;
    MessageLocker::Append("Warning from FieldRangeArray: "
			  "Attempt to step past last step (%d)\n",
			  StepTotal);
    errorcode=-1;
  }
}

double *FieldRangeArray::Field(int index)
{
  static double field[MIFVECDIM];
  if(index<0 || index>StepTotal) index=StepTotal;
  int itemp,irange,extrastep,instep,rangesteps;
  extrastep=instep=rangesteps=0; // Safety
  for(itemp=irange=0;irange<RangeCount;irange++) {
    if(irange>0 && !AreSameField(StopField+MIFVECDIM*(irange-1),
				 StartField+MIFVECDIM*irange))
         extrastep=1;
    else extrastep=0;
    rangesteps=StepCount[irange];
    if(itemp+rangesteps+extrastep>=index) break;
    itemp+=rangesteps+extrastep;
  }
  if(irange>=RangeCount) { // Safety
    irange=RangeCount-1;
    instep=rangesteps=StepCount[irange];
  }
  else  instep=index-itemp-extrastep;

  int i; double c1,c2;
  if(rangesteps<1) { c1=1.0; c2=0.0; }
  else { c2=double(instep)/double(rangesteps);  c1=1.0-c2; }
  if(c1<0.0 || c1>1.0)
    PlainError(1,"Error in FieldRangeArray::Field: %s",ErrProgramming);
  for(i=0;i<MIFVECDIM;i++)
    field[i]=c1*StartField[MIFVECDIM*irange+i]
      +c2*StopField[MIFVECDIM*irange+i];
  return field;
}

double *FieldRangeArray::CurrentField(void)
{
  return Field(CurIndex);
}

FieldRangeArray::~FieldRangeArray(void) { Free(); }

///////////////////// Material specific parameters ///////////////////
// Iron
static const double IRON_Ms= 1700e3;                       //  A/m
static const double IRON_K1=   47e3;                       //  J/m^3
static const double IRON_A=  21e-12;                       //  J/m
static const MifAnisotropyType IRON_AnisType=mif_cubic;

// Nickel
static const double NICKEL_Ms=  485e3;                     //  A/m
static const double NICKEL_K1= -4.5e3;                     //  J/m^3
static const double NICKEL_A=   8e-12;                     //  J/m
static const MifAnisotropyType NICKEL_AnisType=mif_cubic;

// Permalloy
static const double PERMALLOY_Ms=  800e3;                  //  A/m
static const double PERMALLOY_K1=  0.5e3;                  //  J/m^3
static const double PERMALLOY_A=  13e-12;                  //  J/m
static const MifAnisotropyType PERMALLOY_AnisType=mif_uniaxial;

// Cobalt hcp
static const double COBALTHCP_Ms= 1446e3;                  //  A/m
static const double COBALTHCP_K1=  530e3;                  //  J/m^3
static const double COBALTHCP_A=  13e-12;                  //  J/m
static const MifAnisotropyType COBALTHCP_AnisType=mif_uniaxial;

//////////////////////////////////////////////////////////////////////

#define MIFBIGBUFSIZE 4096
#define MIFIDDELIMS ":="                 // Field ID delimiters
#define MIFDELIMS  " \t,;:\"\'`[]{}()="  // Line parameter delimiters
#define MIFCOMMENT '#'     // Character denoting comment start

#define MIFMAXPARAMS 64    // Maximum number of parameters per input line

void MIF::Init(void)
{ // Set MIF member variables to dummy values
  int i;
  CleanUp();
  Ms=A=K1= -1.0;
  AnisType=mif_anis_notset;
  for(i=0;i<MIFVECDIM;i++) AnisDir1[i]=AnisDir2[i]=0.0;
  DemagType=mif_demag_notset;
  PartWidth=PartHeight=PartThickness=CellSize= -1.0;
  GeomType=mif_geom_notset;   GeomTypeParam="0.0";
  RandomizerSeed= -1;
  UserInteractionLevel= -1;
  UserReportCode= -1;
  InitialIncrement= -1.;
  MaxIterationCount=-1;
  AppliedField=(Zeeman *)NULL;
  LogLevel=-1;
}

void MIF::CleanUp(void)
{
  if(AnisInitArgc>0 && AnisInitArgv!=(char **)NULL) {
    delete AnisInitArgv[0]; delete[] AnisInitArgv;
  }
  AnisInitArgc=0; AnisInitArgv=(char **)NULL;
  if(MagInitArgc>0 && MagInitArgv!=(char **)NULL) {
    delete MagInitArgv[0]; delete[] MagInitArgv;
  }
  MagInitArgc=0; MagInitArgv=(char **)NULL;
  FieldRange.FreeSpace();
  InName.Free();
  OutBaseName.Free();
  if(AppliedField!=(Zeeman *)NULL) {
    delete AppliedField; AppliedField=(Zeeman *)NULL;
  }
}

MIF::MIF(void)
{
  AnisInitArgc=0; AnisInitArgv=(char **)NULL;
  MagInitArgc=0; MagInitArgv=(char **)NULL;
  AppliedField=(Zeeman *)NULL;
  SetDefaults();
}

MIF::~MIF(void)
{
  CleanUp();
}

void MIF::SetDefaults(void)
{ // Set some member variables to default values
  int i;
  Init(); // Clear and set variables to dummy values
  for(i=0;i<MIFVECDIM;i++) AnisDir1[i]=AnisDir2[i]=0.0;
  AnisDir1[0]=AnisDir2[1]=1.0;
  AnisInitArgc=1;
  AnisInitArgv=new char*[AnisInitArgc];
  AnisInitArgv[0]=new char[strlen("constant")+1];
  strcpy(AnisInitArgv[0],"constant");
  DemagType=mif_fastpipe;
  GeomType=mif_rectangle;   
  MagInitArgc=1;
  MagInitArgv=new char*[MagInitArgc];
  MagInitArgv[0]=(char *)NULL;
  FieldRange.MakeSpace(1);
  FieldRange.AddRange(); // Zero field, no steps.
  RandomizerSeed=0;
  UserInteractionLevel=0;
  UserReportCode=0;
  LogLevel=0;
  MaxIterationCount=-1; // No limit
  InitialIncrement= 1.0*PI/180.;  // 1 degree.
  const char *cpptr[1]; cpptr[0]="uniform";
  AppliedField=Zeeman::MakeZeeman(1,cpptr);
}

// MIF::ReadLine reads and tokenizes 1 logical line from infile.
// A non-zero return value indicates error; 1 => end-of-file
// The returned string pointers (argp) are only valid until
// the next call to MifReadLine.  Argc should be initialized
// to sizeof(argp)/sizeof(argp[0]) each time before calling this
//  function, as it is used to prevent overflow of argp.
#define MRLsuccess  0  // Successful read
#define MRLeof      1  // Proper end-of-file
#define MRLbadeof   2  // Improper eof
#define MRLferror   3  // Other file error
#define MRLbufof    4  // Input buffer overflow
#define MRLparamof  5  // Parameter space overflow

int MIF::ReadLine(FILE* infile,int &argc,char** argp)
{
  static char buf[MIFBIGBUFSIZE],*cptr;
  // Read one logical line into buf
  int len=0,errflag=0;
  do {
    if(len>=2) len-=2; // Remove trailing backslash and newline
    if(fgets(buf+len,sizeof(buf)-len,infile)==NULL) errflag=1;
    else len+=strlen(buf+len);
  } while(!errflag && len>=2 && buf[len-2]=='\\' && buf[len-1]=='\n');
  if(errflag) { // Unable to read a complete line
    if(len==0 && feof(infile)) return MRLeof;    // Valid eof
    else if(len==0)            return MRLbadeof; // Invalid eof
    else                       return MRLferror; // Other file error
  }
  if(buf[len-1]=='\n') buf[--len]='\0'; // Strip trailing newline
  else if(len==sizeof(buf))  return MRLbufof; // Buffer overflow.
  // Note: The preceding two lines are designed to handle properly the
  // case where the last line of the file doesn't include a trailing
  // newline.

  // Strip comments
  if((cptr=(char *)strchr((const char *)buf,MIFCOMMENT))!=NULL) *cptr='\0';

  // Convert string to standard format
  CollapseStr(buf);
  if(buf[0]=='\0') { // Empty line
    argc=0; argp[0]=(char *)NULL;
    return MRLsuccess;
  }
  LowerString(buf);

  // Tokenize input
  int maxargs=argc;
  argp[argc=0]=strtok(buf,MIFIDDELIMS);  // Field ID
  while(argp[argc]!=NULL) {
    if(argc>=maxargs) return MRLparamof; // Too many arguments
    argp[++argc]=strtok(NULL,MIFDELIMS); // Parameters
  }

  return MRLsuccess;
}

// Fixup AnisDir1 and AnisDir2 as necessary
#define MIFNAME "MIF::AdjustAnisDirs: "
void MIF::AdjustAnisDirs(void)
{
  int i; double temp;
  // Make AnisDir1 a unit vector
  for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=SQ(AnisDir1[i]);
  if(temp<SQRT_EPSILON) PlainError(1,"Error in " MIFNAME "AnisDir1 has zero length");
  if(temp!=1.0) for(i=0,temp=sqrt(temp);i<MIFVECDIM;i++) AnisDir1[i]/=temp;

  // If AnisType is cubic, then AnisDir2 should be unit vector perpendicular
  // to AnisDir1
  if(AnisType==mif_cubic) {
    for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=AnisDir1[i]*AnisDir2[i];
    if(temp!=0.0) { // Eliminate AnisDir1 component from AnisDir2
      for(i=0;i<MIFVECDIM;i++) AnisDir2[i]-=temp*AnisDir1[i];
    }
    // Make AnisDir2 a unit vector
    for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=SQ(AnisDir2[i]);
    if(temp<SQRT_EPSILON) PlainError(1,"Error in " MIFNAME
          "Component of AnisDir2 perpendicular to AnisDir1 has zero length");
    if(temp!=1.0) for(i=0,temp=sqrt(temp);i<MIFVECDIM;i++) AnisDir2[i]/=temp;
  }
}
#undef MIFNAME

#define MIFNAME "MIF::ConsistencyCheck: "
#define MIFERRF(x) PlainWarning("Warning in " MIFNAME "Illegal value for %s: %f",#x,x),\
                    errcount++
#define MIFERRD(x) PlainWarning("Warning in " MIFNAME "Illegal value for %s: %d",#x,x),\
                   errcount++
#define MIFERRE(v,s) PlainWarning("Warning in " MIFNAME "Illegal value for %s: %s",#v,#s),\
                   errcount++
// Routine to perform some consistency checks on member variables
// Return value is the number of errors detected
int MIF::ConsistencyCheck(void)
{
  int i,errcount; double temp;
  errcount=0;
  // Material parameters
  if(Ms<0.0) MIFERRF(Ms);
  if(A<0.0)  MIFERRF(A);
  if(AnisType==mif_anis_notset) MIFERRE(AnisType,mif_anis_notset);
  for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=SQ(AnisDir1[i]);
  if(fabs(temp-1.0)>EPSILON)
    PlainWarning("Warning in " MIFNAME "AnisDir1 has length %f != 1",sqrt(temp)),errcount++;
  if(AnisType==mif_cubic) { // Check for illegal AnisDir2
    for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=SQ(AnisDir2[i]);
    if(fabs(temp-1.0)>EPSILON)
      PlainWarning("Warning in " MIFNAME "AnisDir2 has length %f != 1",sqrt(temp)),errcount++;
    for(i=0,temp=0.0;i<MIFVECDIM;i++) temp+=AnisDir1[i]*AnisDir2[i];
    if(fabs(temp)>EPSILON)
      PlainWarning("Warning in " MIFNAME "AnisDir2 not perpendicular to AnisDir1;"
		  " Inner product =%f",temp),errcount++;
  }
  if(AnisInitList.GetAnisInitFunc(AnisInitArgv[0])==NULL)
    PlainWarning("Warning in " MIFNAME "Unknown AnisInitFunc: ->%s<-",
		 AnisInitArgv[0]),errcount++;

  // Demag specification
  if(DemagType==mif_demag_notset) MIFERRE(DemagType,mif_demag_notset);

  // Part geometry
  if(PartWidth<=0.0)     MIFERRF(PartWidth);
  if(PartHeight<=0.0)    MIFERRF(PartHeight);
  if((DemagType==mif_2dslab || DemagType==mif_3dslab)
     && PartThickness<=0.0) MIFERRF(PartThickness);
  if(CellSize<=0.0)      MIFERRF(CellSize);
  if(CellSize>PartWidth) {
    PlainWarning("Warning in " MIFNAME "CellSize=%g > PartWidth=%g",CellSize,PartWidth);
    errcount++;
  }
  if(CellSize>PartHeight) {
    PlainWarning("Warning in " MIFNAME "CellSize=%g > PartHeight=%g",CellSize,PartHeight);
    errcount++;
  }
  if(GeomType==mif_geom_notset)   MIFERRE(GeomType,mif_geom_notset);


  // Magnetization initialization
  if(MagInit::NameCheck(MagInitArgv[0])==NULL)
    PlainWarning("Warning in " MIFNAME
		 "Unknown magnetization initialization routine %s",
		 MagInitArgv[0]),errcount++;
  else {
    if(MagInit::GetParamCount(MagInitArgv[0])!=MagInitArgc-1)
      PlainWarning("Warning in " MIFNAME
		  "Wrong number of parameters to mag init routine %s,"
		  " %d instead of %d",MagInitArgv[0],MagInitArgc-1,
		  MagInit::GetParamCount(MagInitArgv[0])),errcount++;
  }

  // Experiment parameters
  errcount+=FieldRange.ConsistencyCheck();
  if(AppliedField==(Zeeman *)NULL) {
      PlainWarning("Warning in " MIFNAME "External field type not set");
      errcount++;
  }

  // Output specification
  if(OutBaseName.Length()<1)
    PlainWarning("Warning in " MIFNAME "Output filename not set"),errcount++;
  if(LogLevel<0)
    PlainWarning("Warning in " MIFNAME "Invalid Log level %d<0",LogLevel),errcount++;

  // Miscellaneous
  if(InitialIncrement<=0.0)
    PlainWarning("Warning in " MIFNAME "Initial increment non-positive (%f)",
		InitialIncrement),errcount++;
  if(UserInteractionLevel<0 || UserInteractionLevel>100)
    MIFERRD(UserInteractionLevel);

  return errcount;
}
#undef MIFNAME
#undef MIFERRF
#undef MIFERRD
#undef MIFERRE


#define MIFNAME "MIF::Read: "
#define MIFCOUNT "Incorrect number of parameters (%d) for entry \"%s\""
void MIF::Read(char* inname,int &errorcode)
{ // Returns "".  Argument errorcode == 0 if no error,
  //                                  > 0 on fatal error.
  //                                  < 0 on non-fatal error(s).
  //  Use MessageLocker::GetMessage(String &) to retrieve
  //  error and warning messages (if errorcode!=0).

  int i,paramcount,readcode;
  char *param[MIFMAXPARAMS];

  errorcode=1;     // Cleared after successful completion
  int warncount=0; // Number of non-fatal warnings

  SetDefaults();
  InName=inname;

  FILE* infile;
  if(!(infile=fopen(InName.GetStr(),"r"))) {
    // Add a .mif extension and retry
    InName.Append(".mif");
    if(!(infile=fopen(InName.GetStr(),"r"))) {
      MessageLocker::Append("Error in " MIFNAME "Unable to open file %s\n",
			    inname);
      return;
    }
  }

  do {
    // Read input line
    paramcount=sizeof(param)/sizeof(param[0]);
    if((readcode=ReadLine(infile,paramcount,param))!=MRLsuccess) break;
    if(paramcount<1) continue; // Empty line

    // Process input line; NOTE: All param's are converted to lowercase
    // by ReadLine().

    // MATERIAL PARAMETERS
    if(strcmp(param[0],"base material type")==0 || strcmp(param[0],"bmt")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      //// Hmmm, resetting anisdir's appears to be a bug.
      // for(i=0;i<MIFVECDIM;i++) AnisDir1[i]=AnisDir2[i]=0.0;
      // AnisDir1[0]=AnisDir2[1]=1.0;
      if(strcmp(param[1],"permalloy")==0) {
	Ms=PERMALLOY_Ms; A=PERMALLOY_A; K1=PERMALLOY_K1;
	AnisType=PERMALLOY_AnisType;
      }
      else if(strcmp(param[1],"iron")==0) {
	Ms=IRON_Ms; A=IRON_A; K1=IRON_K1;
	AnisType=IRON_AnisType;
      }
      else if(strcmp(param[1],"nickel")==0) {
	Ms=NICKEL_Ms; A=NICKEL_A; K1=NICKEL_K1;
	AnisType=NICKEL_AnisType;
      }
      else if(strcmp(param[1],"cobalt_hcp")==0 ||
	 strcmp(param[1],"cobalthcp")==0) {
	Ms=COBALTHCP_Ms; A=COBALTHCP_A; K1=COBALTHCP_K1;
	AnisType=COBALTHCP_AnisType;
      }
      else {
	MessageLocker::Append("Error in " MIFNAME "Unknown material: %s\n",
			      param[1]);
	return;
      }
    }
    else if(strcmp(param[0],"ms")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      Ms=atof(param[1]);
    }
    else if(strcmp(param[0],"a")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      A=atof(param[1]);
    }
    else if(strcmp(param[0],"k1")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      K1=atof(param[1]);
    }
    else if(strcmp(param[0],"anisotropy type")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      if(strcmp(param[1],"uniaxial")==0)   AnisType=mif_uniaxial;
      else if(strcmp(param[1],"cubic")==0) AnisType=mif_cubic;
      else {
	MessageLocker::Append("Warning in " MIFNAME
			      "Unknown AnisType: %s\n",param[1]);
	warncount++;
      }
    }
    else if(strcmp(param[0],"anisotropy dir1")==0) {
      if(paramcount!=1+MIFVECDIM) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      for(i=0;i<MIFVECDIM;i++) AnisDir1[i]=atof(param[1+i]);
    }
    else if(strcmp(param[0],"anisotropy dir2")==0) {
      if(paramcount!=1+MIFVECDIM) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      for(i=0;i<MIFVECDIM;i++) AnisDir2[i]=atof(param[1+i]);
    }

    // Anisotropy initialization function
    else if(strcmp(param[0],"anisotropy initialization")==0) {
      if(paramcount<2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      if(AnisInitArgc>0 && AnisInitArgv!=NULL) {
	if(AnisInitArgv[0]!=NULL) delete AnisInitArgv[0];
	delete[] AnisInitArgv;
      }
      AnisInitArgv=new char*[AnisInitArgc=paramcount-1];
      int totalsize=0;
      for(i=0;i<AnisInitArgc;i++) totalsize+=(strlen(param[1+i])+1);
      AnisInitArgv[0]=new char[totalsize];
      for(i=0;i<AnisInitArgc;i++) {
	strcpy(AnisInitArgv[i],param[1+i]);
	if(i+1<AnisInitArgc)
	  AnisInitArgv[i+1]=(char *)strchr((const char *)AnisInitArgv[i],
					   '\0')+1;
      }
    }

    // DEMAG SPECIFICATION
    else if(strcmp(param[0],"demag type")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      if(strcmp(param[1],"slowpipe")==0)      DemagType=mif_slowpipe;
      else if(strcmp(param[1],"fastpipe")==0) DemagType=mif_fastpipe;
      else if(strcmp(param[1],"2dslab")==0)   DemagType=mif_2dslab;
      else if(strcmp(param[1],"3dslab")==0)   DemagType=mif_3dslab;
      else if(strcmp(param[1],"3dcharge")==0) DemagType=mif_3dcharge;
      else if(strcmp(param[1],"none")==0)     DemagType=mif_nodemag;
      else {
	MessageLocker::Append("Warning in " MIFNAME "Unknown DemagType: %s\n",
			      param[1]);
	warncount++;
      }
    }
    // PART GEOMETRY
    else if(strcmp(param[0],"part width")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      PartWidth=atof(param[1]);
    }
    else if(strcmp(param[0],"part height")==0) {
      if(paramcount!=2)
	PlainError(1,"Error in " MIFNAME MIFCOUNT,paramcount-1,param[0]);
      PartHeight=atof(param[1]);
    }
    else if(strcmp(param[0],"part thickness")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      PartThickness=atof(param[1]);
    }
    else if(strcmp(param[0],"cell size")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      CellSize=atof(param[1]);
    }
    else if(strcmp(param[0],"part shape")==0) {
      if(paramcount<2 || paramcount>3) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      if(strcmp(param[1],"rectangle")==0)    GeomType=mif_rectangle;
      else if(strcmp(param[1],"ellipse")==0) GeomType=mif_ellipse;
      else if(strcmp(param[1],"oval")==0)    GeomType=mif_oval;
      else if(strcmp(param[1],"mask")==0)    GeomType=mif_mask;
      else {
	MessageLocker::Append("Warning in " MIFNAME "Unknown GeomType: %s\n",
			      param[1]);
	warncount++;
      }
      if(paramcount==3) GeomTypeParam=param[2];
      else              GeomTypeParam="0.0";
    }
    // INITIAL MAGNETIZATION
    else if(strcmp(param[0],"init mag")==0) {
      if(paramcount<2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      if(MagInitArgc>0 && MagInitArgv!=NULL) {
	if(MagInitArgv[0]!=NULL) delete MagInitArgv[0];
	delete[] MagInitArgv;
      }
      MagInitArgv=new char*[MagInitArgc=paramcount-1];
      int totalsize=0;
      for(i=0;i<MagInitArgc;i++) totalsize+=(strlen(param[1+i])+1);
      MagInitArgv[0]=new char[totalsize];
      for(i=0;i<MagInitArgc;i++) {
	strcpy(MagInitArgv[i],param[1+i]);
	if(i+1<MagInitArgc)
	  MagInitArgv[i+1]=(char *)strchr((const char *)MagInitArgv[i],
					  '\0')+1;
      }
    }
    // EXPERIMENT PARAMETERS
    else if(strcmp(param[0],"field range count")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      FieldRange.MakeSpace(atoi(param[1]));
    }
    else if(strcmp(param[0],"field range")==0) {
      if(paramcount!=2*MIFVECDIM+2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      FieldRange.AddRange(param+1);
    }
    else if(strcmp(param[0],"field type")==0) {
      if(paramcount<2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      if(AppliedField!=(Zeeman *)NULL) delete AppliedField;
      AppliedField=Zeeman::MakeZeeman(paramcount-1,(const char **)(param+1));
    }

    // OUTPUT SPECIFICATION
    else if(strcmp(param[0],"base output filename")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      OutBaseName=param[1];
    }
    else if(strcmp(param[0],"log level")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      LogLevel=atoi(param[1]);
    }
    // MISCELLANEOUS
    else if(strcmp(param[0],"randomizer seed")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      RandomizerSeed=atoi(param[1]);
    }
    else if(strcmp(param[0],"initial increment")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      InitialIncrement=atof(param[1])*PI/180.;
    }
    else if(strcmp(param[0],"user interaction level")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      UserInteractionLevel=atoi(param[1]);
    }
    else if(strcmp(param[0],"user report code")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      UserReportCode=atoi(param[1]);
    }
    else if(strcmp(param[0],"max iteration count")==0) {
      if(paramcount!=2) {
	MessageLocker::Append("Error in " MIFNAME MIFCOUNT "\n",
			      paramcount-1,param[0]);
	return;
      }
      MaxIterationCount=atoi(param[1]);
    }
    // UNKNOWN
    else {
      MessageLocker::Append("Warning in " MIFNAME "Unknown line tag: %s\n",
			    param[0]);
      warncount++;
    }
  } while(readcode==0); // End of input read loop
  fclose(infile);

  // Check for input errors
  if(readcode!=MRLeof) {
    if(readcode==MRLbadeof)
      MessageLocker::Append("Error in " MIFNAME "Improper eof\n");
    else if(readcode==MRLferror)
      MessageLocker::Append("Error in " MIFNAME "File error\n");
    else if(readcode==MRLbufof)
      MessageLocker::Append("Error in " MIFNAME "Input buffer overflow\n");
    else if(readcode==MRLparamof)
      MessageLocker::Append("Error in " MIFNAME "Parameter space overflow\n");
    else 
      MessageLocker::Append("Error in " MIFNAME "Unknown file input error\n");
    return;
  }

  // Adjust AnisDir1 and AnisDir2
  AdjustAnisDirs();

  // Check for valid specification
  int errcount;
  if((errcount=ConsistencyCheck())!=0) {
    MessageLocker::Append("Error in " MIFNAME 
			  "%d error%s found in input file %s\n",
			  errcount,(errcount==1?"":"s"),inname);
    return;
  }

  // If we get here, then there are no fatal errors.
  errorcode = -warncount;  // Possibly some warnings, though.
  return;
}
#undef MIFNAME
#undef MIFCOUNT

const char* MIF::GetDemagName(void)
{
  const char* cptr;
  switch(DemagType)
    {
    case mif_slowpipe:  cptr=MifSlowPipeName; break;
    case mif_fastpipe:  cptr=MifFastPipeName; break;
    case mif_2dslab:    cptr=Mif2dSlabName;   break;
    case mif_3dslab:    cptr=Mif3dSlabName;   break;
    case mif_3dcharge:  cptr=Mif3dchargeName;   break;
    case mif_nodemag:   cptr=MifNoDemagName;  break;
    default:            cptr=MifFastPipeName; break;
    }
  return cptr;
}

const double* MIF::GetAnisDir(int i)
{
  if(i==1) return AnisDir1;
  if(i==2) return AnisDir2;
  return (const double*)NULL;
}

void MIF::DumpPartialInfo(FILE* fptr)
{
  int i;
  fprintf(fptr,"# MIF input file: %s\n",InName.GetStr());
  fprintf(fptr,"#  Material information---\n");
  fprintf(fptr,"#   Ms=%.5e  A=%.5e  K1=%.5e\n",Ms,A,K1);
  fprintf(fptr,"#   AnisType= %s\n",
	  (AnisType==mif_uniaxial ? "Uniaxial" :
	   (AnisType==mif_cubic ? "Cubic" : "Not set")));
  fprintf(fptr,"#   AnisDir1=(%6.3f,%6.3f,%6.3f)\n",
	  AnisDir1[0],AnisDir1[1],AnisDir1[2]);
  if(AnisType==mif_cubic)
    fprintf(fptr,"#   AnisDir2=(%6.3f,%6.3f,%6.3f)\n",
	    AnisDir2[0],AnisDir2[1],AnisDir2[2]);
  fprintf(fptr,"#   AnisInit=");
  for(i=0;i<AnisInitArgc;i++) fprintf(fptr," %s",AnisInitArgv[i]);
  fprintf(fptr,"\n");
  fprintf(fptr,"#  Demag type: ");
  switch(DemagType)
    {
    case mif_slowpipe:     fprintf(fptr,"Slow Pipe\n"); break;
    case mif_fastpipe:     fprintf(fptr,"Fast Pipe\n"); break;
    case mif_2dslab:       fprintf(fptr,"2D Slab\n"); break;
    case mif_3dslab:       fprintf(fptr,"3D Slab\n"); break;
    case mif_3dcharge:       fprintf(fptr,"3D Charge\n"); break;
    case mif_nodemag:      fprintf(fptr,"No demag\n"); break;
    case mif_demag_notset: fprintf(fptr,"Not set\n"); break;
    default:           fprintf(fptr,"ERROR: UNKNOWN DEMAG TYPE\n"); break;
    }
  fprintf(fptr,"#  Geometry information---\n");
  fprintf(fptr,"#   Part width:     %.4e m\n",PartWidth);
  fprintf(fptr,"#   Part height:    %.4e m\n",PartHeight);
  fprintf(fptr,"#   Part thickness: %.4e m\n",PartThickness);
  fprintf(fptr,"#   Cell size:      %.4e m\n",CellSize);
  fprintf(fptr,"#   Part shape:     %s %s\n",
	  (GeomType==mif_rectangle ? "rectangle" :
	   (GeomType==mif_ellipse ? "ellipse" :
	   (GeomType==mif_mask ? "Mask" : 
	    (GeomType==mif_oval ? "oval" : "not set")))),GeomTypeParam.GetStr());
  fprintf(fptr,"#  Initial Magnetization:");
  for(i=0;i<MagInitArgc;i++) fprintf(fptr," %s",MagInitArgv[i]);
  fprintf(fptr,"\n");
  fprintf(fptr,"#  Randomizer seed: %d\n",RandomizerSeed);
  fprintf(fptr,"#  Output base name: %s\n",OutBaseName.GetStr());
  fflush(fptr);
}

