# FILE: array.tcl
#
# Since the member variables of a class may only be simple variables,
# we must provide a mechanism for simulating array variables.
# This is it.  
# 
# It makes use of the internal implementation of em_class, so major
# changes to em_class may require changes here.  This breaks the
# encapsulation rules, but is necessary here to provide the array 
# functionality.
#
# TODO:
# Add methods that imitate remaining subcommands of Tcl's 
# 'array' command: startsearch, anymore, nextelement, donesearch
#
# Modify 'print' method to pretty print, like parray does

# Only define em_array once (bootstrap protection)
if {[string match {} [info procs em_array]]} {

em_class em_array {

  method set {arglist} {} {
    if {[llength $arglist] % 2} {
      error "List argument must have even number of elements

usage:
	\$instance set list

Invalid call:
	\$instance set $arglist"
    } else {
      set quotedlist {}
      foreach {name value} $arglist {
        lappend quotedlist @$name $value
      }
      upvar #0 $this state
      array set state $quotedlist
    }
  }

  method append {name value} {} {
    upvar #0 $this state
      append state(@$name) $value
    return $state(@$name)
  }

  method exists { name } {} {
    upvar #0 $this state
    return [info exists state(@$name)]
  }

  method get {{pattern *}} {} {
    upvar #0 $this state
    set quotedlist [array get state @$pattern]
    set plainlist {}
    foreach {quotedname value} $quotedlist {
      regexp @(.*) $quotedname match name
      lappend plainlist $name $value
    }
    return $plainlist
  }
  
  method unset {name} {} {
    upvar #0 $this state
    if {[catch {unset state(@$name)} msg]} {
      error "can't unset element '$name' in em_array '$this': no such element"
    } else {
      return $msg
    }
  }

  method names {{pattern *}} {} {
    upvar #0 $this state
    set quotednames [array names state @$pattern]
    set plainnames {}
    foreach quotedname $quotednames {
      regexp @(.*) $quotedname match name
      lappend plainnames $name 
    }
    return $plainnames
  }

  method size {} {} {
    upvar #0 $this state
    llength [array names state @*]
  }

  method lookup {name} {} {
    upvar #0 $this state
    if {[catch {set state(@$name)} value]} {
      error "can't access element '$name' in em_array '$this': no such element"
    }
    return $value
  }

  method print {} {} {
    foreach name [$this names] {
      puts "${this}($name) = [$this lookup $name]"
    }
  }

}

}
