# FILE: link.tcl
#
# An object which implements an event-driven, asynchronous, line-based
# network connection.  It generates an event Readable, when a line
# is available for reading, and queues lines for eventual writing.
# This is the object which actually owns and manages a Tcl socket.
#
# The Tcl event loop must be running for this event-driven object to function.
#

em_class net_link {

    # A maximum backlog on the queue, for detecting possible bugs
    common maxQueueLength 100

    # The socket through which we communicate to the host.
    const public variable socket

    # The name of the remote host
    public variable hostname 

    # The remote port
    public variable port 
    
    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler delete
        }
        set logHandler [em_eventHandler new $log delete "$this delete"]
        if {$imadelog} {
            $oldlog delete
        }
        set oldlog $log
    }
    private variable oldlog = em_nop
    private variable logHandler
    private variable imadelog = 0

    # line read from socket
    private variable lineFromSocket

    # queue of messages to write to socket
    private variable putQueue = {}

    # My binding tags
    private variable bindtags

    private variable timeoutevent

    # Constructor is a non-blocking operation.  If it doesn't return
    # an error, it will generate a Ready event if and when it has established
    # a socket connection.  Then, it will generate Readable events whenever
    # it has a line to be read (using the Get method).  It will generate a
    # delete event whenever it is destroyed.
    constructor { args } {
        eval $this configure $args
        if {![info exists log]} {
            # Provide a default log
            $this configure -log [em_log new -title "Connection $this Log"]
            set imadelog 1
        }
        # Enable event generation
        set bindtags [list $this $class]
        # Try to open connection to host
        if {![info exists socket]} {
            if {![info exists hostname] || ![info exists port]} {
                error "Missing required options: -hostname and -port"
            }
            if {[catch {socket -async $hostname $port} msg]} {
                $log Append "Can't connect to $hostname:$port:
	$msg"
                error "Can't connect to $hostname:$port:
	$msg"
            }
            set socket $msg
            # Grrrr.  Win95 doesn't appear to set the socket writable
            # on failure.  Have to time out.
$log Append "Hostname = $hostname"
            if {[string match localhost $hostname]} {
$log Append "Timeout in 1000 ms"
                set timeoutevent [after 1000 $this SocketOpen]
            } else {
$log Append "Timeout in 120000 ms"
                set timeoutevent [after 120000 $this SocketOpen]
            }
        }
        fileevent $socket writable [list $this SocketOpen]
    }

    # handler called when connection to server is established (or not)
    method SocketOpen {} {
        if {[info exists timeoutevent]} {
            after cancel $timeoutevent
            unset timeoutevent
        }
        fileevent $socket writable {}
        if {[catch {fconfigure $socket -peername} peerinfo]} {
            $log Append "No answer on $hostname:$port"
            $this delete
            return
        }
        set hostname [lindex $peerinfo 1]
        set port [lindex $peerinfo 2]
        fconfigure $socket -blocking 0 -buffering line -translation {auto crlf}
        $log Append "Connected to $hostname:$port"
        $log Append "Peer: $peerinfo"
        em_eventHandler Generate Ready $bindtags
        fileevent $socket readable [list $this Receive]
    }

    # handler called when there is data available from the host
    # It invokes other events describing what it receives.
    method Receive {} {
        if {[catch {gets $socket line} readCount]} {
            $this SocketError $readCount
            return
        }
        if {$readCount < 0} {
            if {[fblocked $socket]} {
                # 'gets' call is blocked until a whole line is available
                return
            }
            if {[eof $socket]} {
                $this SocketEOF
                return
            }
        }
        regsub -all {%([0-9a-fA-F][0-9a-fA-F])} $line \
                {[scan \1 %x v; format %c $v]} line
        if {[catch {subst $line} lineFromSocket]} {
            $log Append "Receive subst error: $lineFromSocket
	Data: $line"
            return
        }
        em_eventHandler Generate Readable $bindtags
    }

    private method SocketError { msg } {
        $log Append "socket error in connection $this to $hostname:$port:
	$msg"
        $this delete
        return
    }

    private method SocketEOF {} {
        $log Append "socket closed by $hostname:$port in $this"
        $this delete
        return
    }

    method Get {} {
        return $lineFromSocket
    }

    method Put { string } {
        # First escape all % signs, so we know from now on any %'s we see came
        # from escaping.
        regsub -all % $string %25 string
        # Then escape characters which need special treament because of the
        # way the Tcl interpreter parses them
        regsub -all \{ $string %7b string
        regsub -all {\[} $string %5b string
        regsub -all \" $string %22 string
        regsub -all {\\} $string %5c string
        regsub -all "\n" $string %0a string
        # Escape everything else that we like.  Add to the list.
        regsub -all "\[\$\]" $string \
                {%[scan & %c v; format %02x $v]} string
        if {[catch {subst $string} sstring]} {
            $log Append "Put subst error: $sstring
	Data: $string"
            return
        }
        $this Queue $sstring
    }
 
    private method Queue { msg } {
        if {![llength $putQueue]} {
            fileevent $socket writable [list $this Send]
        }
        lappend putQueue $msg
        if {[llength $putQueue] > $maxQueueLength} {
            # will this message ever be seen?
            $log Append "Maximum queue length exceeded in link $this:
	Possibly a socket error?"
            $this delete
            return
        }
    }

    method Send {} {
        # This won't block, but could leave data backed up in an internal
        # Tcl buffer.  That's unlikely since we're waiting for a writable
        # socket.
        puts $socket [lindex $putQueue 0]
        set putQueue [lrange $putQueue 1 end]
        if {![llength $putQueue]} {
            fileevent $socket writable {}
        }
    }

    destructor {
        if {[info exists bindtags]} {
            em_eventHandler Generate delete $bindtags
        }
        if {[info exists timeoutevent]} {
            after cancel $timeoutevent
        }
        if {[info exists socket]} {
            fileevent $socket readable {}
            # Flush the queue
            while {[llength $putQueue]} {
                $this Send
            }
            $log Append "Closing socket $socket in $this"
            close $socket
        }
        if {[info exists log]} {
            $log Append "Deleting $this ..."
            $this configure -log em_nop
        }
    }

}

