# FILE: solver.tcl
#
# Class for controlling micromagnetic solver
#

em_class omf_solver {
    public variable mif_filename {
        if {[info exists mif]} {
            $mif delete
        }
        if {[info exists grid]} {
            $grid delete
        }
	set mif [mms_mif new]
	$mif Read $mif_filename
	set grid [mms_grid2D new $mif]
	$this Reset
    }
    private variable mif
    private variable grid
    public variable relax_stop_torque = 1e-5
    public variable relax_stop_proc = "$this RelaxStopCheck"
    public variable stop_request_flag = 0
    ## The RelaxState method performs an update between
    ## ODE solver steps, after which it checks this variable
    ## for an outside stop request.
    private variable ode_step_count = 0  ;# Total number of ODE
    ## steps taken.  This is only incremented (never reset) during
    ## the life of the instance.
    private variable solver_running = 0 ;# Set true while a
    ## calculation is ongoing.  Prevents more than one solving
    ## thread from being started.

    private variable bindtags

    constructor { args} {
	eval $this configure $args
        set bindtags [list $this $class]
    }
    method Reset {} {
	$mif ResetField
	$grid ResetODE
	set stop_request_flag 0
	set solver_running 0
    }
    method GetIterationValueLabels {} {
        return \
"  ODE     Max       Max                               ENERGIES
 Step    Torque    Angle     Total       Exchange    Anisotropy      Demag        Zeeman"
    }
    method GetIterationValues {} {
	set stepcount [$this GetODEStepCount]
	set torque [$this GetMaxTorque]
	set maxangle [$this GetMaxAngle]
	set energies [join [$this GetEnergies]]
	return [format \
		"%5d %10.7f  %5.1f  %12.8f %12.8f %12.8f %12.8f %12.8f" \
		$stepcount $torque $maxangle \
		[lindex $energies 1] [lindex $energies 3] \
		[lindex $energies 5] [lindex $energies 7] \
		[lindex $energies 9]]
    }
    method RelaxStopCheck { {torque unknown} } {
	# Returns 1 if stopping criteria are met, 0 otherwise.
	# Call with torque set to "unknown" (the default)
	# to query $grid for MaxTorque.
	if { [string match "unknown" $torque] } {
	    # Query grid for max torque
	    set torque [$grid GetMaxTorque]
	}
	if {[expr $torque<$relax_stop_torque]} {
	    return 1      ;# STOP!
	}
	return 0
    }
    method StepSolver {} { grid } {
	set torque [$grid StepODE]
	return $torque
    }
    method RelaxState { {relax_callback em_nop} } {
	# The command "relax_callback" is executed before returning
	# if $relax_stop_proc is satisfied.
	if {$solver_running} { return -1 } ;# Solver already running
	set solver_running 1     ;# Set solver semaphore
	set stop_request_flag 0  ;# Reset stop flag
	$grid ResetODE
	set torque [$grid GetMaxTorque]
	while { 1 } {
            em_eventHandler Generate Iteration $bindtags
	    if {[eval $relax_stop_proc $torque]} {
		# Relaxed state reached
		eval $relax_callback
		break
	    }

	    # Process background processes and check
	    # for outside stop request.
	    update
	    if { "$stop_request_flag" != "0" } { break }

	    # Step ODE solver
	    set torque [$grid StepODE]
	    incr ode_step_count
	}
	set solver_running 0     ;# Clear solver semaphore
	return $torque
    }
    method HystLoop { {relax_callback em_nop} \
	    { fieldstep_callback em_nop } } {
	# The command "relax_callback is passed on to RelaxState.
	# The command "callback" is executed after each field
	# step, at which time the calling routine may query the
	# solver for information and do any desired housekeeping.
	if {$solver_running} { return -1 } ;# Solver already running
	while { 1 } {
            em_eventHandler Generate FieldStep $bindtags
	    $this RelaxState $relax_callback

	    # Check for outside stop request
	    if { "$stop_request_flag" != "0" } { break }

	    if {[$mif GetFieldStepIndex]<[$mif GetFieldStepTotal]} {
		$mif StepField
		eval $fieldstep_callback
	    } else {
		break
	    }
	}
    }
    method IncrementFieldStep { { amount 1 }} {
	set index [$mif GetFieldStepIndex]
	incr index $amount
	if {$amount<0} {set $amount 0}
	set newindex [$mif SetFieldStepIndex $index]
	$grid ResetODE
	return $newindex
    }
    method GetFieldStepCount {} {
        return [$mif GetFieldStepIndex]
    }
    method GetODEStepCount {} { return $ode_step_count; }
    method GetMaxTorque {} {
	return [$grid GetMaxTorque]
    }
    method GetMaxAngle {} {
	# Returns max neighbor angle in degrees.
	return [expr [$grid GetMaxNeighborAngle]*180./3.14159265358979]
    }
    method GetEnergies {} {
	# Returns a list of list pairs, energy name + value
	# The first is the total energy"
	set energy_list [$grid GetEnergies]
	set rtn_list {}
	lappend rtn_list [list "Total" [lindex $energy_list 4]]
	lappend rtn_list [list "Exchange" [lindex $energy_list 0]]
	lappend rtn_list [list "Anisotropy" [lindex $energy_list 1]]
	lappend rtn_list [list "Demag" [lindex $energy_list 2]]
	lappend rtn_list [list "Zeeman" [lindex $energy_list 3]]
	return $rtn_list
    }
    method GetNomAppliedField { {mult 1.} {fmt "%.10g %.10g %.10g"}} {
	set field [$grid GetNomBApplied]
	# Note: 2D solver uses "z" up and "y" into page,
	#  as opposed to the rest of the OOMMF project
	#  that uses "y" up and "z" out of the page.
	return [format $fmt \
		[expr [lindex $field 0]*$mult] \
		[expr [lindex $field 2]*$mult] \
		[expr -1*[lindex $field 1]*$mult]]
    }
    method GuessAppliedFieldAxis {} {
	#  Queries the mif file and tries to guess the component
	# of BApplied that is most likely intended for M-B loop.
	#  Returns a unit 3 vector.
	#  See note in GetNomAppliedField about required coordinate
	# system transform.
	set B0 [$mif GetNomExternalField 0]
	set B1 [$mif GetNomExternalField 1]
	set dBx [expr abs([lindex $B1 0] - [lindex $B0 0])]
	set dBy [expr abs([lindex $B1 2] - [lindex $B0 2])]
	set dBz [expr abs([lindex $B0 1] - [lindex $B1 1])]
	if { $dBz > $dBy && $dBz > $dBx } { return "0 0 1" }
	if { $dBy > $dBx }                { return "0 1 0" }
	return "1 0 0"
    }
    method GetAveMag {} {
	# Returns average magnetization M/Ms, as a string with
	# 3 entries.  Same notes apply as in GetNomAppliedField
	set mag [$grid GetAveMag]
	return "[lindex $mag 0] [lindex $mag 2] \
		[expr -1*[lindex $mag 1]]"
    }
    method GetOutputBasename {} {
	return [$mif GetOutBaseName]
    }
    method WriteFieldFile { {filename {}} {note {}} } {
	# Writes reduced h field to a vio file
	if {[string match {} $filename]} {
	    # Use data in mif to construct filename
	    set filename [format "%s.pass%05dh.vio" \
		    [$mif GetOutBaseName] [$this GetODEStepCount]]
	}
	$grid BinDumph $filename $note
        return [file join [pwd] $filename]
    }
    method WriteMagFile { {filename {}} {note {}} } {
	# Writes magnetization to a vio file
	if {[string match {} $filename]} {
	    # Use data in mif to construct filename
	    set filename [format "%s.pass%05dm.vio" \
		    [$mif GetOutBaseName] [$this GetODEStepCount]]
	}
	$grid BinDumpm $filename $note
        return [file join [pwd] $filename]
    }
    destructor {
        em_eventHandler Generate delete $bindtags
        if {[info exists grid]} {
	    $grid delete
        }
        if {[info exists mif]} {
	    $mif delete
        }
    }
}

